@@ -470,7 +470,7 @@ class HeterozygousSite:
         
         return self.aligned_het_bases
     
-    def calculate_probs(self, haplotype: object = None) -> float:
+    def calculate_probs(self, haplotype: object = None, force_recalc = False) -> float:
         """
         Args:
             haplotype (object):
@@ -478,6 +478,9 @@ class HeterozygousSite:
         Returns:
             float:
         """
+        # Use cached probs_read if possible.
+        if len(self.probs_read) > 0 and not force_recalc:
+            return sum(self.probs_read)
         
         total_non_matches: int = 0
         self.total_matches = 0
@@ -491,7 +494,9 @@ class HeterozygousSite:
             haplotype = haplotype
         
         var = self.automated_evaluation_of_local_alignments
-        probs_read: List[Union[float, Any]] = []
+        
+        self.probs_read: List[Union[float, Any]] = []
+        
         lengths: List[Any] = []
         for pos in self.sorted_pos:
             if pos[0][0] == str(haplotype):
