@@ -1,14 +1,20 @@
 # coding=utf-8
+import sys
 import math
 from collections import defaultdict
 from statistics import mean
 from typing import Any, List
 import pysam
+import numpy as np
 
 import LRphase.HeterozygousSites as HeterozygousSite
 
+# Minimum floating-point value > 0
+# See http://www.seunosewa.com/2009/06/floating-point-min-max.html
+minfloat = 4.9406564584124654e-324
+log10minfloat = math.log10(minfloat)
 
-def powlaw_modified(x: object, a: object = 4, xmin: object = 2) -> object:
+def powlaw_modified(x: object, a: object = 4.5, xmin: object = 2) -> object:
     """
 
     Args:
@@ -22,10 +28,13 @@ def powlaw_modified(x: object, a: object = 4, xmin: object = 2) -> object:
     return ((a - 1) / xmin) * math.pow(x / xmin, -1 * a)
 
 
+"""
 def _multcoeff(*args):
-    '''# self.result = _multcoeff(*args)'''
-    """Return the multinomial coefficient
-    (n1 + n2 + ...)! / n1! / n2! / ..."""
+    '''
+    # self.result = _multcoeff(*args)
+    Return the multinomial coefficient
+    (n1 + n2 + ...)! / n1! / n2! / ...
+    '''
     if not args:  # no parameters
         return 1
     # Find and store the index of the largest parameter so we can skip
@@ -40,7 +49,24 @@ def _multcoeff(*args):
             result = result * num // den
             num += 1
     return result
+"""
 
+# Cache multinomial coefficients for efficiency
+_multcoeff_cache = {}
+
+def _multcoeff(n_match, n_mismatch):
+    """
+    Calculates multinomial coefficient according to the formula
+    n! / ((n-m)! * m!)  (As described in the LRphase article)
+    where n = the total number of heterozygous sites in a region 
+    (n_match + n_mismatch) and m = the number of heterozygous sites
+    matching the given haplotype (n_match). Values are cached to
+    improve performance.
+    """
+    if (n_match, n_mismatch) not in _multcoeff_cache.keys():
+        _multcoeff_cache[(n_match, n_mismatch)] = math.log10(np.math.factorial(n_match+n_mismatch) / (np.math.factorial(n_mismatch) * np.math.factorial(n_match)))
+    return _multcoeff_cache[(n_match, n_mismatch)]
+    
 
 def _read_base_error_rates_from_quality_scores(gapped_alignment_positions, query_qualities):
     """# self.read_base_error_rates, self.error_rate_average_het_sites =
@@ -74,6 +100,8 @@ def _read_bases_from_gapped_alignment(gapped_alignment_positions, query_sequence
     self.aligned_segment.query_sequence) """
     ############ Extract bases for heterozygous sites in aligned_segment  ############
     read_bases = []
+    if query_sequence is None:
+        return read_bases
     for position in gapped_alignment_positions:
         if isinstance(position, str):
             if position[0] == '-':
@@ -224,18 +252,39 @@ def _count_matches(
     total_hets = 0
     total_hets_analyzed = 0
     ploidy_list = []
-    i = 0
+
     # print(read_bases, phased_alleles, read_base_error_rates, error_model)
-    for read_base in read_bases:
+    for i in range(len(read_bases)):
+        read_base = read_bases[i]
         ploidy = len(phased_alleles[i])
         ploidy_list.append(ploidy)
         q = read_base_error_rates[i] + .001
-        if not read_base == '-':
+        if not read_base == '-': # Skip gap positions
+            """
+            The loop below does not work consistently. In its current state, it is inconsistent
+            in how complex variants are handled (i.e., those where one or more alleles is >1bp)
+            in that these are counted as mismatches if the allele for the present haplotype is
+            a single bp, but are skipped entirely when the current allele has len > 1. This means
+            that log-likelihoods are calculated using different numbers of heterozygous sites per
+            haplotype, which is incorrect. This commit adds functionality to filter complex variants
+            out of the variant list prior to even running this step. This is done by default when the
+            PhaseSet object is initialized. Given this change in behavior, the current implementation
+            of this loop produces correct and consistent scores. However, this should be modified in
+            future releases to calculate correct scores given all types of potential variants. In
+            short, I'm leaving this as-is for now and relying on only simple SNPs being passed in
+            in phased_alleles.
+            """
             for haplotype in range(0, ploidy):
-                if len(phased_alleles[i][haplotype]) > 1:
+                # Begin relocated block
+                if len(phased_alleles[i][haplotype]) > 1:  # Probably to exclude indels/rearrangements
                     continue
-            total_hets_analyzed += 1
-            for haplotype in range(0, ploidy):
+                total_hets_analyzed += 0.5  # These are counted twice
+                # end relocated block
+                """
+                It is necessary to log-transform to avoid underflows. However, there may
+                be a faster way to apply this that does not involve taking logs at every
+                iteration!
+                """
                 if read_base == phased_alleles[i][haplotype]:
                     log_probability_read_given_haplotype_i[haplotype] += math.log10(1 - q)
                     matches[haplotype] += 1
@@ -243,7 +292,13 @@ def _count_matches(
                     log_probability_read_given_haplotype_i[haplotype] += math.log10(q / 3)
                     non_matches[haplotype] += 1
         total_hets += 1
-        i += 1
+    """
+    sys.stderr.write("Total hets: %d, total_analyzed: %d, matches[0]: %d, mismatches[0]: %d, matches[1], %d, mismatches[1]: %d, q: %.5f\n" %
+                     (total_hets, total_hets_analyzed,
+                      matches[0], non_matches[0],
+                      matches[1], non_matches[1], q)
+    )
+    """
     if total_hets_analyzed > 0:
         ploidy_phase_set = max(ploidy_list)
     else:
@@ -252,10 +307,10 @@ def _count_matches(
     if multinomial_correction:
         haplotype: int
         for haplotype in range(0, ploidy_phase_set):
-            multinomial = math.log10(_multcoeff(int(matches[haplotype]), int(non_matches[haplotype])))
+            multinomial = _multcoeff(int(matches[haplotype]), int(non_matches[haplotype]))
             log_probability_read_given_haplotype_i[haplotype] = log_probability_read_given_haplotype_i[
                                                                     haplotype] + multinomial
-    
+            #sys.stderr.write("Haplotype %d log_prob: %.5f, prior: %.5f\n" % (haplotype, log_probability_read_given_haplotype_i[haplotype], multinomial))
     return log_probability_read_given_haplotype_i, non_matches, matches, total_hets, total_hets_analyzed, ploidy_phase_set
 
 
@@ -285,14 +340,58 @@ def _estimate_per_base_error_rate(aligned_segment: pysam.AlignedSegment) -> obje
     return per_base_error_rate, error_rate_calculation
 
 
-def _get_phased_alleles(vcf_records, sample):
+def _get_phased_alleles(vcf_records, sample, use_complex_variants = True, use_gap_variants = True):
     '''self.phased_alleles = self._get_phased_alleles(self.vcf_records, self.sample) '''
-    return [vcf_record.samples[sample].alleles for vcf_record in vcf_records]
+    provisional_recs = [vcf_record.samples[sample].alleles for vcf_record in vcf_records]
+
+    if use_complex_variants and use_gap_variants:
+        return provisional_recs
+    elif len(provisional_recs) == 0:
+        return provisional_recs
+
+    haplotypes = range(len(provisional_recs[0]))
+    ret = []
+    for rec in provisional_recs:
+        use_rec = True
+        for haplotype in haplotypes:
+            if not use_complex_variants and len(rec[haplotype]) > 1: # Drop complex variants
+                use_rec = False
+                break
+            if not use_gap_variants and rec[haplotype] == '-': # Drop gap variants
+                use_rec = False
+                break
+        if use_rec:
+            ret.append(rec)
+
+    return ret
 
 
-def _get_reference_positions(vcf_records):
+def _get_reference_positions(vcf_records, sample,  use_complex_variants = True, use_gap_variants = True):
     '''self.reference_positions = self._get_reference_positions(self.vcf_records) '''
-    return [vcf_record.pos - 1 for vcf_record in vcf_records]
+    provisional_pos = [vcf_record.pos - 1 for vcf_record in vcf_records] # Converts pos to zero-based coord. frame
+    provisional_recs = [vcf_record.samples[sample].alleles for vcf_record in vcf_records]
+
+    if use_complex_variants and use_gap_variants:
+        return provisional_pos
+    elif len(provisional_recs) == 0:
+        return provisional_pos
+
+    haplotypes = range(len(provisional_recs[0]))
+    ret = []
+    for i in range(len(provisional_recs)):
+        rec = provisional_recs[i]
+        use_rec = True
+        for haplotype in haplotypes:
+            if not use_complex_variants and len(rec[haplotype]) > 1:
+                use_rec = False
+                break
+            if not use_gap_variants and rec[haplotype] == '-':
+                use_rec = False
+                break
+        if use_rec:
+            ret.append(provisional_pos[i])
+
+    return ret
 
 
 def _generate_gapped_alignment(reference_start, cigartuples, reference_positions) -> object:
@@ -341,9 +440,12 @@ def _generate_gapped_alignment(reference_start, cigartuples, reference_positions
 
 
 def _estimate_prior_probabilities(ploidy_phase_set = 2, prior_probabilities: object = None):
+    """
+    Returns log10 transformed priors.
+    """
     if prior_probabilities is None:
         if ploidy_phase_set > 0:
-            prior_probabilities = [1 / ploidy_phase_set] * ploidy_phase_set  # P(phase_i)
+            prior_probabilities = list(map(math.log10, [1 / ploidy_phase_set] * ploidy_phase_set))  # P(phase_i)
     return prior_probabilities
 
 
@@ -354,7 +456,8 @@ class PhaseSet:
     
     def __init__(
             self, phase_set, aligned_segment, vcf_records, sample, error_model = 0, error_rate_threshold = 0.01,
-            multinomial_correction = True, prior_probabilities = None, liftover_converters = None
+            multinomial_correction = True, prior_probabilities = None, liftover_converters = None, phase = 'Nonphasable',
+            powlaw_alpha = 4.5, powlaw_xmin = 2, use_complex_variants = False, use_gap_variants = False
             ):
         self.aligned_segment = aligned_segment
         self.phase_set = phase_set
@@ -367,7 +470,7 @@ class PhaseSet:
         self.sample = sample
         self.error_model = error_model
         self.error_rate_threshold = error_rate_threshold
-        self.prior_probabilities = _estimate_prior_probabilities()
+        self.prior_probabilities = _estimate_prior_probabilities()  # Log10 transformed priors
         self.multinomial_correction = multinomial_correction
         self.gapped_alignment_positions = []
         self.read_bases = []
@@ -375,11 +478,21 @@ class PhaseSet:
         self.reference_positions = []
         self.max_log_likelihood_ratio = None
         self.max_phase = None
-        self.phase = None
-        if not self.aligned_segment.is_unmapped or self.aligned_segment.is_aligned_to_contig_not_in_vcf:
+        self.phase = phase
+        self.powlaw_alpha = powlaw_alpha
+        self.powlaw_xmin = powlaw_xmin
+        
+        self.use_complex_variants = use_complex_variants
+        self.use_gap_variants = use_gap_variants
+        
+        self.total_hets = 0
+        if not self.aligned_segment.is_unmapped: # or self.aligned_segment.is_aligned_to_contig_not_in_vcf:
+            # Don't think second test is necessary because it should never yield len(self.vcf_records) > 0.
+            # Current implementation of PhasedRead object sometimes leaves this attribute unset, causing
+            # exceptions, so better if we can avoid it.
             if len(self.vcf_records) > 0:
-                self.reference_positions = _get_reference_positions(self.vcf_records)
-                self.phased_alleles = _get_phased_alleles(self.vcf_records, self.sample)
+                self.reference_positions = _get_reference_positions(self.vcf_records, self.sample, self.use_complex_variants, self.use_gap_variants)
+                self.phased_alleles = _get_phased_alleles(self.vcf_records, self.sample, self.use_complex_variants, self.use_gap_variants)
                 self.gapped_alignment_positions = _generate_gapped_alignment(
                         self.aligned_segment.reference_start,
                         self.aligned_segment.cigartuples,
@@ -390,7 +503,8 @@ class PhaseSet:
                         self.aligned_segment.query_sequence
                         )
         # self.read_base_error_rates =[]
-    
+
+        
     def __iter__(self):
         self.HeterozygousSites = []
         self.HeterozygousSites_processed_count = 0
@@ -439,7 +553,7 @@ class PhaseSet:
                     
                     self.log_likelihood_ratios, self.phase, self.max_phase, self.max_log_likelihood_ratio = _phasing_decision(
                             self.log_probability_read_given_haplotype_i, self.ploidy_phase_set, prior_probabilities,
-                            error_rate_threshold, self.total_hets_analyzed
+                            error_rate_threshold, self.total_hets_analyzed, powlaw_alpha=self.powlaw_alpha, powlaw_xmin=self.powlaw_xmin
                             )
                 
                 elif error_model == 5:
@@ -454,7 +568,7 @@ class PhaseSet:
                     for haplotype in range(1, self.ploidy_phase_set + 1):
                         haplotypes[haplotype] = sum(
                                 [het.calculate_probs(haplotype = haplotype) for het in hets if
-                                 het.calculate_probs(haplotype = haplotype)]
+                                 het.calculate_probs(haplotype = haplotype)] # Second call to calculate_probs uses cached values
                                 )
                         # if len(haplotypes[haplotype])>0:
                         # print(haplotypes[haplotype])
@@ -556,35 +670,12 @@ class PhaseSet:
                                 )
         
         return read_base_error_rates, error_rate_average_het_sites, per_base_mismatch_rate
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
-    
-    # @staticmethod
 
 
 def _phasing_decision(
         log_probability_read_given_haplotype_i, ploidy_phase_set, prior_probabilities,
-        error_rate_threshold, total_hets_analyzed
+        error_rate_threshold, total_hets_analyzed, powlaw_alpha = 4.5, powlaw_xmin = 2,
+        minfloat = log10minfloat
         ) -> object:
     """# self.log_likelihood_ratios, self.phase, self.max_phase, self.max_log_likelihood_ratio =
     self._phasing_decision(self.log_probability_read_given_haplotype_i, self.ploidy_phase_set,
@@ -607,58 +698,46 @@ def _phasing_decision(
     else:
         log_likelihood_ratios = []
         for haplotype in log_probability_read_given_haplotype_i:
-            log_flag = False
             denominator = 0
             numerator = 0
             for j in range(0, ploidy_phase_set):
                 if haplotype == j:
+                    # Assumes priors are log-transformed! 
+                    numerator += (log_probability_read_given_haplotype_i[haplotype] +
+                                  prior_probabilities[haplotype])
+                    """
                     try:
-                        numerator += math.log10(
-                                (10**log_probability_read_given_haplotype_i[haplotype]) * prior_probabilities[
-                                    haplotype]
-                                )
+                        # Assumes priors are log-transformed!
+                        numerator += (log_probability_read_given_haplotype_i[haplotype] +
+                                      prior_probabilities[haplotype])
                     except ValueError:
                         numerator += log_probability_read_given_haplotype_i[haplotype]
+                    """
                 else:
-                    if log_flag:
-                        if denominator == 0:
-                            denominator = log_probability_read_given_haplotype_i[j]
-                        elif denominator < 0 and denominator < log_probability_read_given_haplotype_i[j]:
-                            denominator = log_probability_read_given_haplotype_i[j]
-                    try:
-                        test: float = math.log10(
-                                (10**log_probability_read_given_haplotype_i[j]) * prior_probabilities[j]
-                                )
-                        denominator += (10**log_probability_read_given_haplotype_i[j]) * prior_probabilities[j]
-                    except ValueError:
-                        if denominator == 0:
-                            denominator = log_probability_read_given_haplotype_i[j]
-                            log_flag = True
-                        elif denominator < 0 and denominator < log_probability_read_given_haplotype_i[j]:
-                            denominator = log_probability_read_given_haplotype_i[j]
-                            log_flag = True
-                        elif denominator > 0:
-                            if math.log10(denominator) < log_probability_read_given_haplotype_i[j]:
-                                denominator = log_probability_read_given_haplotype_i[j]
-                                log_flag = True
-            if log_flag:
-                log_likelihood_ratios.append(numerator - denominator)
+                    # Assumes priors are log-transformed! 
+                    denominator += 10**(log_probability_read_given_haplotype_i[j] + prior_probabilities[j])
+            #sys.stderr.write("hap: %d, numer: %f, denom %f\n" % (haplotype, numerator, denominator))
+                    
+            # This heuristic is needed to account for edge cases in which a read
+            # has a very large number of heterozygous sites, most of which match
+            # a single phase. These can cause underflows in the denominator.
+            try:
+                denom = math.log10(denominator)
+            except ValueError:
+                denom = log10minfloat
+            log_likelihood_ratios.append(numerator - denom)
+
+        #sys.stderr.write("%s\n" % (log_likelihood_ratios))
+        max_log_likelihood_ratio = max(log_likelihood_ratios)
+        if max_log_likelihood_ratio == 0:
+            phase = 'Unphased'
+            max_phase = 0
+        else:
+            max_phase = log_likelihood_ratios.index(max_log_likelihood_ratio) + 1
+            if powlaw_modified(max_log_likelihood_ratio, a=powlaw_alpha, xmin=powlaw_xmin) <= error_rate_threshold:
+                # if max_log_likelihood_ratio >= 10 ** (math.log10((error_rate_threshold / .5491)) / -2.45):
+                phase = max_phase
             else:
-                log_likelihood_ratios.append(numerator - math.log10(denominator))
-        i = 0
-        for log_likelihood_ratio in log_likelihood_ratios:
-            if max(log_likelihood_ratios) == 0:
                 phase = 'Unphased'
-                max_phase = 0
-                max_log_likelihood_ratio = 0
-            elif log_likelihood_ratio == max(log_likelihood_ratios):
-                max_log_likelihood_ratio = log_likelihood_ratios[i]
-                max_phase = i + 1
-                if error_rate_threshold >= powlaw_modified(max_log_likelihood_ratio):
-                    # if max_log_likelihood_ratio >= 10 ** (math.log10((error_rate_threshold / .5491)) / -2.45):
-                    phase = i + 1
-                else:
-                    phase = f'Unphased'
-            i += 1
         
         return [log_likelihood_ratios, phase, max_phase, max_log_likelihood_ratio]
